#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from pydantic_xml import BaseXmlModel, element, attr
from ..constants import WSDL_NSMAP


__all__ = [
    "WSDLFaultPort",
    "WSDLOutputPort",
    "WSDLInputPort",
    "WSDLDocumentation",
    "WSDLOperationPort",
    "WSDLPortType"
]


class WSDLFaultPort(BaseXmlModel, tag="fault", nsmap=WSDL_NSMAP):
    message: str = attr()
    name: str = attr()


class WSDLOutputPort(BaseXmlModel, tag="output", nsmap=WSDL_NSMAP):
    message: str = attr()
    name: str = attr()


class WSDLInputPort(BaseXmlModel, tag="input", nsmap=WSDL_NSMAP):
    message: str = attr()
    name: str = attr()


class WSDLDocumentation(BaseXmlModel, tag="documentation", nsmap=WSDL_NSMAP):
    title: str = element(ns='xroad', nsmap=WSDL_NSMAP)
    notes: str = element(ns='xroad', nsmap=WSDL_NSMAP)


class WSDLOperationPort(BaseXmlModel, tag="operation", ns="wsdl", nsmap=WSDL_NSMAP):
    documentation: WSDLDocumentation
    input: WSDLInputPort
    output: WSDLOutputPort
    fault: WSDLFaultPort
    name: str = attr()


class WSDLPortType(BaseXmlModel, tag="portType", ns="wsdl", nsmap=WSDL_NSMAP):
    name: str = attr(default="servicePortType")
    operations: t.List[WSDLOperationPort]



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
from __future__ import annotations
import base64
import string
import secrets
import typing as t
from weakref import WeakValueDictionary


__all__ = [
	"GlobalWeakStorage",
	"ExportableGlobalWeakStorage"
]


class GlobalWeakStorage:
	_instances = WeakValueDictionary()
	_inst_counter = 0
	_uid = ''

	def __new__(cls) -> GlobalWeakStorage:
		uid = cls._unique_id(cls._inst_counter)
		instance = super().__new__(cls)
		instance._uid = uid
		cls._instances[uid] = instance
		cls._inst_counter += 1
		return instance

	def __init__(self) -> None:
		"""
		The GlobalWeakStorage class is a global repository for storing weak object
		references within its instances. This design allows Python to automatically
		garbage collect unused objects that are stored in GlobalWeakStorage, but
		have lost all strong references in the running program. Objects are inserted
		into and stored in instances of this class, but they can also be globally
		retrieved by their unique identifiers using GlobalWeakStorage class methods.
		"""
		self._objects = WeakValueDictionary()
		self._obj_counter = 0

	@staticmethod
	def _unique_id(index: int) -> str:
		"""Generates a unique identifier."""
		clamped = index % 1_000_000_000
		filled = str(clamped).zfill(9)
		token = secrets.token_bytes(36)
		b64_token = base64.b64encode(token)
		return f"{filled}..{b64_token.decode()}"

	@staticmethod
	def validate_fingerprint(fingerprint: str, raise_on_invalid: bool = True) -> bool:
		"""Validates the format of unique identifiers."""
		valid_b64_chars = string.ascii_letters + string.digits + '+/='
		is_valid = True
		if not isinstance(fingerprint, str):
			is_valid = False
		elif len(fingerprint) != 122:
			is_valid = False
		elif '-$$-' not in fingerprint:
			is_valid = False
		if is_valid:
			for uid in fingerprint.split('-$$-'):
				if len(uid) != 59:
					is_valid = False
				elif '..' not in uid:
					is_valid = False
				counter, token = uid.split('..')  # type: str, str
				if len(counter) != 9:
					is_valid = False
				elif not counter.isdigit():
					is_valid = False
				for char in token:
					if char not in valid_b64_chars:
						is_valid = False
		if not is_valid and raise_on_invalid:
			raise ValueError(f"invalid fingerprint: {fingerprint}")
		return is_valid

	def insert_object(self, obj: t.Any) -> str:
		"""
		Inserts an object into the GlobalWeakStorage, generating a unique identifier for it.

		:param obj: The object to be stored in this GlobalWeakStorage instance.
		:returns: A unique identifier for the object, which can be used to access the
			object from any instance of the GlobalWeakStorage class or the class itself.
		"""
		uid = self._unique_id(self._obj_counter)
		self._obj_counter += 1
		self._objects[uid] = obj
		return f"{self._uid}-$$-{uid}"

	@classmethod
	def retrieve_object(cls, fingerprint: str, *, raise_on_miss: bool = True) -> t.Union[t.Any, None]:
		"""
		Retrieves an object from an instance of GlobalWeakStorage identified by the provided fingerprint.

		:param fingerprint: The unique identifier of the object.
		:param raise_on_miss: Whether a KeyError should be raised if the
			fingerprint does not relate to any object. Defaults to True.
		:raises ValueError: If the fingerprint has invalid format.
		:raises KeyError: if `raise_on_miss` is True and the fingerprint does not relate to any object.
		:returns: The object associated with the given fingerprint if it exists, otherwise None.
		"""
		cls.validate_fingerprint(fingerprint)
		inst_id, obj_id = fingerprint.split('-$$-')
		instance = cls._instances.get(inst_id) or {}
		store: WeakValueDictionary = getattr(instance, '_objects', {})
		if raise_on_miss:
			return store[obj_id]
		return store.get(obj_id, None)

	def get(self, fingerprint: str, *, raise_on_miss: bool = True) -> t.Union[t.Any, None]:
		"""
		Retrieves an object from this instance of GlobalWeakStorage identified by the provided fingerprint.

		:param fingerprint: The unique identifier of the object.
		:param raise_on_miss: Whether a KeyError should be raised if the
			fingerprint does not relate to any object. Defaults to True.
		:raises ValueError: If the fingerprint has invalid format.
		:raises KeyError: if `raise_on_miss` is True and the fingerprint does not relate to any object.
		:returns: The object associated with the given fingerprint if it exists, otherwise None.
		"""
		self.validate_fingerprint(fingerprint)
		obj_id = fingerprint.split('-$$-')[1]
		if raise_on_miss:
			return self._objects[obj_id]
		return self._objects.get(obj_id, None)


ExportableGlobalWeakStorage = type('GlobalWeakStorage', (GlobalWeakStorage,), {})



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
from pydantic_xml import BaseXmlModel, attr
from ..constants import WSDL_NSMAP


__all__ = [
	"SOAPAddress",
	"WSDLPortBinding",
	"WSDLService"
]


class SOAPAddress(BaseXmlModel, tag="address", ns="soap", nsmap=WSDL_NSMAP):
	location: str = attr(default="http://example.org/xroad-endpoint")


class WSDLPortBinding(BaseXmlModel, tag="port", ns="wsdl", nsmap=WSDL_NSMAP):
	binding: str = attr(default="tns:serviceBinding")
	name: str = attr(default="servicePort")
	address: SOAPAddress = SOAPAddress()


class WSDLService(BaseXmlModel, tag="service", nsmap=WSDL_NSMAP):
	name: str = attr(default="xroadService")
	port: WSDLPortBinding = WSDLPortBinding()



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from email import policy
from email import encoders
from email.mime.text import MIMEText
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from ..elements import SwaRefInternal


__all__ = ["MultipartEncoder"]


class MultipartEncoder:
	def __init__(self, content: bytes, files: t.List[SwaRefInternal]) -> None:
		multipart = self.compute_multipart(content, files)
		headers, message = multipart.split(b"\r\n\r\n", 1)
		self.headers = self.raw_headers_to_dict(headers)
		self.message = b'\r\n' + message  # type: bytes
		self.headers["Content-Length"] = str(len(self.message))

	@classmethod
	def compute_multipart(cls, xml_str: bytes, files: t.List[SwaRefInternal]) -> bytes:
		root = MIMEMultipart("related", type="text/xml", start="<rootpart>")
		cls.attach_soap_envelope(xml_str, root)
		if len(files) > 1:
			mixed = MIMEMultipart("mixed")
			mixed.add_header("Content-Transfer-Encoding", "binary")
			for file in files:
				cls.attach_file(file, mixed)
			root.attach(mixed)
		else:
			cls.attach_file(files[0], root)
		return root.as_bytes(policy=policy.HTTP)

	@staticmethod
	def attach_soap_envelope(xml_str: bytes, parent: MIMEMultipart):
		part = MIMEText(xml_str.decode('utf-8'))
		part.replace_header("Content-Transfer-Encoding", "8bit")
		part.replace_header("Content-Type", "text/xml; charset=UTF-8")
		part.add_header("Content-ID", "<rootpart>")
		parent.attach(part)

	@staticmethod
	def attach_file(file: SwaRefInternal, parent: MIMEMultipart):
		part = MIMEBase(*file.mimetype.split('/'))
		part.add_header("Content-ID", file.mime_cid)
		part.add_header("Content-Digest", file.digest)
		part.add_header("Content-Length", str(file.size))
		part.add_header('Content-Disposition', '; '.join([
			"attachment", f'name="{file.name}"', f'filename="{file.name}"'
		]))
		part.set_payload(file.content)
		encoders.encode_base64(part)
		parent.attach(part)

	@staticmethod
	def raw_headers_to_dict(raw_headers: bytes) -> t.Dict[str, str]:
		headers = dict()
		parts = raw_headers.split(b"\r\n")
		for part in parts:
			key, value = part.decode().split(': ')
			headers[key] = value
		return headers



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from .bodypart import DecodedBodyPart
from .errors import NonMultipartError
from .. import utils


__all__ = ["MultipartDecoder"]


class MultipartDecoder:
    def __init__(self, content, content_type) -> None:
        self.content_type = content_type
        self.parts: t.Tuple[DecodedBodyPart, ...] = tuple()
        self._find_boundary()
        self._parse_body(content)

    def _find_boundary(self):
        ct_info = tuple(x.strip() for x in self.content_type.split(';'))
        mimetype = ct_info[0]
        if mimetype.split('/')[0].lower() != 'multipart':
            raise NonMultipartError(mimetype)
        for item in ct_info[1:]:
            attr, value = utils.split_on_find(item, separator='=')
            if attr.lower() == 'boundary':
                string = value.strip('"')
                if not (string is None or isinstance(string, bytes)):
                    self.boundary = string.encode('utf-8')
                    continue
                self.boundary = string

    @staticmethod
    def _fix_first_part(part, boundary_marker):
        bm_len = len(boundary_marker)
        if boundary_marker == part[:bm_len]:
            return part[bm_len:]
        return part

    def _parse_body(self, content: bytes) -> None:
        boundary = b''.join((b'--', self.boundary))

        def body_part(part):
            fixed = MultipartDecoder._fix_first_part(part, boundary)
            return DecodedBodyPart(fixed)

        def test_part(part):
            return (part != b'' and
                    part != b'\r\n' and
                    part[:4] != b'--\r\n' and
                    part != b'--')

        parts = content.split(b''.join((b'\r\n', boundary)))
        self.parts = tuple(body_part(x) for x in parts if test_part(x))



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from pydantic_xml import BaseXmlModel, attr


__all__ = [
	"Enumeration",
	"RegexPattern",
	"MinInclusive",
	"MaxInclusive",
	"MinExclusive",
	"MaxExclusive",
	"FractionDigits",
	"TotalDigits",
	"Length",
	"MinLength",
	"MaxLength",
	"WhiteSpace"
]


class Enumeration(BaseXmlModel, tag="enumeration"):
	value: str = attr()


class RegexPattern(BaseXmlModel, tag="pattern"):
	value: str = attr()


class MinInclusive(BaseXmlModel, tag="minInclusive"):
	value: str = attr()


class MaxInclusive(BaseXmlModel, tag="maxInclusive"):
	value: str = attr()


class MinExclusive(BaseXmlModel, tag="minExclusive"):
	value: str = attr()


class MaxExclusive(BaseXmlModel, tag="maxExclusive"):
	value: str = attr()


class FractionDigits(BaseXmlModel, tag="fractionDigits"):
	value: str = attr()


class TotalDigits(BaseXmlModel, tag="totalDigits"):
	value: str = attr()


class Length(BaseXmlModel, tag="length"):
	value: str = attr()


class MinLength(BaseXmlModel, tag="minLength"):
	value: str = attr()


class MaxLength(BaseXmlModel, tag="maxLength"):
	value: str = attr()


class WhiteSpace(BaseXmlModel, tag="whiteSpace"):
	value: t.Literal["preserve", "replace", "collapse"] = attr(default="preserve")



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from pydantic_xml import BaseXmlModel, attr
from ..constants import WSDL_NSMAP


__all__ = [
	"SOAPFault",
	"SOAPBody",
	"SOAPHeader",
	"WSDLFaultBinding",
	"WSDLOutputBinding",
	"WSDLInputBinding",
	"XROADVersion",
	"SOAPOperationBinding",
	"WSDLOperationBinding",
	"SOAPBinding",
	"WSDLBinding"
]


class SOAPFault(BaseXmlModel, tag="fault", ns="soap", nsmap=WSDL_NSMAP):
	name: str = attr(default="FaultResponse")
	use: str = attr(default="literal")


class SOAPBody(BaseXmlModel, tag="body", ns="soap", nsmap=WSDL_NSMAP):
	use: str = attr(default="literal")


class SOAPHeader(BaseXmlModel, tag="header", ns="soap", nsmap=WSDL_NSMAP):
	message: str = attr(default="tns:xroadHeader")
	use: str = attr(default="literal")
	part: str = attr()


class WSDLFaultBinding(BaseXmlModel, tag="fault", ns="wsdl", nsmap=WSDL_NSMAP):
	name: str = attr(default="FaultResponse")
	fault: SOAPFault = SOAPFault()


class WSDLOutputBinding(BaseXmlModel, tag="output", ns="wsdl", nsmap=WSDL_NSMAP):
	headers: t.List[SOAPHeader] = [
		SOAPHeader(part="client"),
		SOAPHeader(part="service"),
		SOAPHeader(part="id"),
		SOAPHeader(part="userId"),
		SOAPHeader(part="protocolVersion")
	]
	body: SOAPBody = SOAPBody()


class WSDLInputBinding(BaseXmlModel, tag="input", ns="wsdl", nsmap=WSDL_NSMAP):
	headers: t.List[SOAPHeader] = [
		SOAPHeader(part="client"),
		SOAPHeader(part="service"),
		SOAPHeader(part="id"),
		SOAPHeader(part="protocolVersion"),
		SOAPHeader(part="userId")
	]
	body: SOAPBody = SOAPBody()


class XROADVersion(BaseXmlModel, tag="version", ns="xroad", nsmap=WSDL_NSMAP):
	version: str = "v1"


class SOAPOperationBinding(BaseXmlModel, tag="operation", ns="soap", nsmap=WSDL_NSMAP):
	soap_action: str = attr(name="soapAction")


class WSDLOperationBinding(BaseXmlModel, tag="operation", ns="wsdl", nsmap=WSDL_NSMAP):
	name: str = attr()
	operation: SOAPOperationBinding
	version: XROADVersion = XROADVersion()
	input: WSDLInputBinding = WSDLInputBinding()
	output: WSDLOutputBinding = WSDLOutputBinding()
	fault: WSDLFaultBinding = WSDLFaultBinding()


class SOAPBinding(BaseXmlModel, tag="binding", ns="soap", nsmap=WSDL_NSMAP):
	style: str = attr(default="document")
	transport: str = attr(default="http://schemas.xmlsoap.org/soap/http")


class WSDLBinding(BaseXmlModel, tag="binding", ns="wsdl", nsmap=WSDL_NSMAP):
	name: str = attr(default="serviceBinding")
	type: str = attr(default="tns:servicePortType")
	binding: SOAPBinding
	operations: t.List[WSDLOperationBinding]



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from fastapi import Response
from ..base import MessageBody
from ..cid_gen import CIDGenerator
from ..multipart import MultipartEncoder
from ..elements import (
	SwaRefInternal,
	SwaRefSpec
)
from ..envelope import (
	EnvelopeFactory,
	XroadHeader
)


__all__ = ["SoapResponse"]


FileObject = t.Union[SwaRefInternal, t.List[SwaRefInternal]]


class SoapResponse(Response):
	media_type = 'text/xml'

	def __init__(
			self,
			content: MessageBody,
			header: t.Optional[XroadHeader] = None,
			http_status_code: t.Optional[int] = 200
	) -> None:
		if files := self.assign_content_ids(content):
			message = self.serialize(content, header)
			encoder = MultipartEncoder(message, files)
			xml_str = encoder.message
			http_headers = encoder.headers
		else:
			xml_str = self.serialize(content, header)
			http_headers = {
				"Content-Type": "text/xml;charset=UTF-8"
			}
		super().__init__(
			content=xml_str,
			headers=http_headers,
			status_code=http_status_code
		)

	@classmethod
	def assign_content_ids(cls, content: MessageBody, gen: CIDGenerator = None) -> t.List[SwaRefInternal]:
		gen, files = gen or CIDGenerator(), list()
		if not isinstance(content, MessageBody):
			return []

		# Define recursion behavior
		for sub_content in vars(content).values():
			if isinstance(sub_content, MessageBody):
				nested = cls.assign_content_ids(sub_content, gen)
				files.extend(nested)

		# Add files from content
		specs = getattr(content, "_element_specs", None)
		if specs is None:
			return files
		for attr, spec in specs.items():
			if not isinstance(spec, SwaRefSpec):
				continue
			obj: FileObject = getattr(content, attr)
			if isinstance(obj, SwaRefInternal):
				obj.content_id = gen.token
				files.append(obj)
			elif isinstance(obj, list):
				for item in obj:
					if isinstance(item, SwaRefInternal):
						item.content_id = gen.token
						files.append(item)
		return files

	@staticmethod
	def serialize(content: MessageBody, header: t.Optional[XroadHeader]) -> bytes:
		envelope = EnvelopeFactory[content.__class__]()
		return envelope.serialize(
			content=content,
			header=header,
			pretty_print=False,
			skip_empty=True
		)



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
from pydantic_xml import model
from ..constants import A8nType
from .spec import BaseElementSpec


__all__ = ["ElementSpecMeta", "CompositeMeta"]


class ElementSpecMeta(type):
	def __new__(cls, name, bases, class_fields, **kwargs):
		class_a8ns: dict = class_fields.get("__annotations__", {})
		element_specs = dict()
		for attr, obj in class_fields.items():
			if isinstance(obj, BaseElementSpec):
				a8n = class_a8ns.get(attr, A8nType.ABSENT)
				obj.set_a8n_type_from(a8n, attr, name)
				class_fields[attr] = obj.get_element(attr)
				class_a8ns[attr] = obj.get_element_a8n()
				element_specs[attr] = obj
		class_fields["__annotations__"] = class_a8ns
		class_fields["_element_specs"] = element_specs
		return super().__new__(cls, name, bases, class_fields, **kwargs)


class CompositeMeta(ElementSpecMeta, model.XmlModelMeta):
	pass



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from ..base import BaseElementSpec
from .validators import CommonValidators, NumberValidators, StringValidators


__all__ = ["CommonSpecTypeA", "CommonSpecTypeB"]


class CommonSpecTypeA(BaseElementSpec, CommonValidators, NumberValidators):
	def __init__(self, element_type: t.Any, **kwargs) -> None:
		self.min_value = kwargs.pop("min_value")
		self.max_value = kwargs.pop("max_value")
		self.enumerations = kwargs.pop("enumerations")
		self.pattern = kwargs.pop("pattern")
		super().__init__(
			element_type=element_type,
			**kwargs
		)

	def process(self, obj: t.Any) -> t.Any:
		self.validate_pattern(obj)
		self.validate_total_digits(obj)
		self.validate_min_max_value(obj)
		self.validate_enumerations(obj)
		return obj

	def init_instantiated_data(self, data: t.List[t.Any]) -> t.List[t.Any]:
		return [self.process(obj) for obj in data]

	def init_deserialized_data(self, data: t.List[t.Any]) -> t.List[t.Any]:
		return [self.process(obj) for obj in data]


class CommonSpecTypeB(BaseElementSpec, CommonValidators, StringValidators):
	def __init__(self, element_type: t.Any, **kwargs) -> None:
		self.length = kwargs.pop("length")
		self.min_length = kwargs.pop("min_length")
		self.max_length = kwargs.pop("max_length")
		self.enumerations = kwargs.pop("enumerations")
		self.pattern = kwargs.pop("pattern")
		self.whitespace = kwargs.pop("whitespace")
		super().__init__(
			element_type=element_type,
			**kwargs
		)

	def process(self, obj: t.Any) -> t.Any:
		self.validate_string_length(obj)
		self.validate_pattern(obj)
		self.validate_enumerations(obj)
		return self.process_whitespace(obj)

	def init_instantiated_data(self, data: t.List[t.Any]) -> t.List[t.Any]:
		return [self.process(obj) for obj in data]

	def init_deserialized_data(self, data: t.List[t.Any]) -> t.List[t.Any]:
		return [self.process(obj) for obj in data]



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from pydantic_xml import BaseXmlModel, attr
from ..constants import WSDL_NSMAP
from .restrictions import (
    StringRestriction,
    IntegerRestriction,
    DecimalRestriction,
    FloatRestriction,
    DoubleRestriction,
    DateRestriction,
    TimeRestriction,
    DateTimeRestriction,
    DurationRestriction,
    AnyURIRestriction
)


__all__ = [
    "SimpleType",
    "AnyXML",
    "Element",
    "Sequence",
    "ComplexType",
    "Import",
    "Include",
    "Schema"
]


class SimpleType(BaseXmlModel, tag="simpleType"):
    name: str = attr()
    restriction: t.Union[
        StringRestriction,
        IntegerRestriction,
        DecimalRestriction,
        FloatRestriction,
        DoubleRestriction,
        DateRestriction,
        TimeRestriction,
        DateTimeRestriction,
        DurationRestriction,
        AnyURIRestriction
    ]


class AnyXML(BaseXmlModel, tag="any"):
    process_contents: str = attr(default="lax")


class Element(BaseXmlModel, tag="element"):
    name: str = attr()
    type: str = attr()
    min_occurs: t.Union[str, None] = attr(name="minOccurs", default=None)
    max_occurs: t.Union[str, None] = attr(name="maxOccurs", default=None)


class Sequence(BaseXmlModel, tag="sequence"):
    elements: t.List[t.Union[Element, AnyXML]]


class ComplexType(BaseXmlModel, tag="complexType"):
    name: str = attr()
    sequence: Sequence


class Import(BaseXmlModel, tag="import"):
    schema_loc: str = attr(name="schemaLocation")
    namespace: str = attr()


class Include(BaseXmlModel, tag="include"):
    schema_loc: str = attr(name="schemaLocation")


class Schema(BaseXmlModel, tag="schema"):
    xmlns: str = attr(default=WSDL_NSMAP["xsd"])
    target_ns: str = attr(name="targetNamespace")
    imports: t.List[Import]
    includes: t.List[Include]
    elements: t.List[Element]
    complex_types: t.List[ComplexType]
    simple_types: t.List[SimpleType]



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
from __future__ import annotations
import base64
import hashlib
import typing as t
from pydantic import Field
from ..base import BaseElementSpec, MessageBody
from ..storage import GlobalWeakStorage
from ..multipart import DecodedBodyPart
from ..file_size import FileSize
from .. import utils


__all__ = [
	"SwaRefFile",
	"SwaRefInternal",
	"SwaRefSpec",
	"SwaRefElement",
	"SwaRef"
]


_NSMap = t.Optional[t.Dict[str, str]]
_FileType = t.Union[DecodedBodyPart, None]
_FileSizeType = t.Union[FileSize, None]
_Filetypes = t.Union[t.List[str], t.Literal["all"]]
_HashFuncType = t.Literal["sha256", "sha512", "sha3_256", "sha3_384", "sha3_512"]


class SwaRefFile(MessageBody):
	name: str = Field(exclude=True, default='')
	size: int = Field(exclude=True, default=0)
	digest: str = Field(exclude=True, default='')
	mimetype: str = Field(exclude=True, default='')
	content: bytes = Field(exclude=True, default='')

	def __new__(cls, name: str, content: bytes) -> SwaRefFile:
		kwargs = {k: v for k, v in locals().items() if v != cls}
		return t.cast(SwaRefFile, SwaRefInternal(**kwargs))

	@classmethod
	def _real_new_(cls, sub_cls):
		return super().__new__(sub_cls)


class SwaRefInternal(SwaRefFile):
	content_id: str = Field(default=None)

	def __new__(cls, **__):
		return super()._real_new_(cls)

	def __init__(self, **kwargs):
		if len(kwargs) == 1:
			key, value = list(kwargs.items())[0]
			if key == "name":
				kwargs = dict(content_id=value)
		super().__init__(**kwargs)

	@property
	def mime_cid(self) -> str:
		cid = self.content_id
		if isinstance(cid, str) and ':' in cid:
			raw = self.content_id.split(':')[1]
			return f"<{raw}>"
		return ''


class SwaRefSpec(BaseElementSpec):
	def __init__(self, **kwargs) -> None:
		self.allowed_filetypes = kwargs.pop("allowed_filetypes")
		self.max_filesize = kwargs.pop("max_filesize")
		self.hash_func = kwargs.pop("hash_func")
		super().__init__(
			element_type=SwaRefFile,
			internal_type=SwaRefInternal,
			**kwargs
		)

	def digest(self, content: bytes) -> str:
		hash_func = getattr(hashlib, self.hash_func)
		digest = hash_func(content).digest()
		name = self.hash_func.replace('_', '-')
		b64_hash = base64.b64encode(digest).decode()
		return f"{name}={b64_hash}"

	def validate_file(self, name: str, size: int, content_id: str = None) -> None:
		extract = '' if content_id is None else f"$${content_id}$$"
		if '.' not in name:
			raise ValueError(f"invalid file name: {name}")
		if self.allowed_filetypes != "all":
			file_ext = '.' + name.split('.')[-1]
			if file_ext not in self.allowed_filetypes:
				raise ValueError(
					f"file type not allowed: {file_ext} "
					f"(allowed: {self.allowed_filetypes}){extract}"
				)
		if self.max_filesize and size > self.max_filesize.value:
			size_str = FileSize.bytes_to_iec_str(size)
			raise ValueError(
				f"file size too large: {size_str} "
				f"(max: {self.max_filesize}){extract}"
			)

	def init_instantiated_data(self, data: t.List[SwaRefInternal]) -> t.List[SwaRefInternal]:
		for obj in data:
			obj.size = len(obj.content)
			self.validate_file(obj.name, obj.size)

			obj.digest = self.digest(obj.content)
			obj.mimetype = utils.guess_mime_type(obj.name)
			obj.content = utils.convert_to_utf8(obj.content)

			if hasattr(obj, "content_id"):
				delattr(obj, "content_id")
		return data

	def init_deserialized_data(self, data: t.List[SwaRefInternal]) -> t.List[SwaRefInternal]:
		for obj in data:
			if not hasattr(obj, "content_id"):
				continue
			try:
				file: DecodedBodyPart = GlobalWeakStorage.retrieve_object(obj.content_id)
			except ValueError:
				raise ValueError(f"no file attachment found $${obj.content_id}$$")

			obj.name = file.file_name
			obj.size = len(file.content)
			self.validate_file(obj.name, obj.size, file.content_id)

			obj.digest = self.digest(file.content)
			obj.mimetype = utils.guess_mime_type(file.file_name)
			obj.content = utils.convert_to_utf8(file.content)

			setattr(obj, "_file", file)
			delattr(obj, "content_id")
		return data


class SwaRefElement:
	def __new__(
			cls,
			*,
			tag: t.Optional[str] = None,
			ns: t.Optional[str] = None,
			nsmap: _NSMap = None,
			min_occurs: int = None,
			max_occurs: t.Union[int, t.Literal["unbounded"]] = None,
			allowed_filetypes: _Filetypes = "all",
			max_filesize: _FileSizeType = None,
			hash_func: _HashFuncType = "sha3_512"
	) -> t.Union[SwaRefFile, t.List[SwaRefFile]]:
		kwargs = {k: v for k, v in locals().items() if v != cls}
		return t.cast(SwaRefFile, SwaRefSpec(**kwargs))


class SwaRef:
	File: t.Type[SwaRefFile] = SwaRefFile
	Element: t.Type[SwaRefElement] = SwaRefElement

	def __init__(self):
		raise NotImplementedError(
			"Cannot directly instantiate SwaRef, you must "
			"instantiate one of its attribute classes."
		)



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
from .factory import (
	EnvelopeFactory
)
from .generics import (
	GenericEnvelope,
	GenericFault,
	GenericBody,
	AnyBody
)
from .header import (
	XroadService,
	XroadClient,
	XroadHeader
)


__all__ = [
	"EnvelopeFactory",
	"GenericEnvelope",
	"GenericFault",
	"GenericBody",
	"AnyBody",
	"XroadService",
	"XroadClient",
	"XroadHeader"
]



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
import inflection
from abc import ABC, abstractmethod
from pydantic_xml import model, element
from ..constants import A8nType

try:
	from .body import MessageBody
except ImportError:  # pragma: no cover
	MessageBody: t.TypeAlias = t.Any


__all__ = ["BaseElementSpec"]


class BaseElementSpec(ABC):
	tag: t.Optional[str]
	ns: t.Optional[str]
	nsmap: t.Optional[t.Dict[str, str]]
	min_occurs: int
	max_occurs: t.Union[int, t.Literal["unbounded"]]
	element_type: t.Type[MessageBody]
	internal_type: t.Optional[t.Type[MessageBody]]
	a8n_type: t.Union[A8nType, None]

	def __init__(
			self,
			element_type: t.Type[MessageBody],
			internal_type: t.Optional[t.Type[MessageBody]] = None,
			**kwargs
	) -> None:
		self.tag = kwargs.get("tag")
		self.ns = kwargs.get("ns")
		self.nsmap = kwargs.get("nsmap")
		self.min_occurs = kwargs.get("min_occurs") or 0
		self.max_occurs = kwargs.get("max_occurs") or "unbounded"
		self.element_type = element_type
		self.internal_type = internal_type
		self.a8n_type = None

	@abstractmethod
	def init_instantiated_data(self, data: t.List[t.Any]) -> t.List[t.Any]: ...

	@abstractmethod
	def init_deserialized_data(self, data: t.List[t.Any]) -> t.List[t.Any]: ...

	def get_element_a8n(self) -> t.Type[t.List[t.Any]]:
		return t.List[self.internal_type or self.element_type]

	def get_element(self, attr: str) -> model.XmlEntityInfo:
		return element(
			tag=self.tag or inflection.camelize(attr),
			ns=self.ns or '',
			nsmap=self.nsmap or dict(),
			default_factory=list
		)

	def set_a8n_type_from(self, a8n: t.Any, attr: str, cls_name: str) -> None:
		if a8n in [A8nType.ABSENT, self.element_type]:
			self.a8n_type = A8nType.MAND
			return
		elif a8n in [t.Optional, t.Union, t.List, list]:
			raise TypeError(
				f"Not permitted to provide '{a8n}' "
				f"annotation without a type argument."
			)
		if args := t.get_args(a8n):
			if len(args) == 1 and args[0] != self.element_type:
				raise TypeError(
					f"Single annotation argument for class '{cls_name}' "
					f"attribute '{attr}' must be '{self.element_type.__name__}'."
				)
			for arg in args:
				if arg is type(None):
					continue
				elif arg == self.element_type:
					continue
				arg_name = "None" if arg is type(None) else arg.__name__
				raise TypeError(
					f"Invalid annotation argument '{arg_name}' "
					f"for '{cls_name}' class attribute '{attr}'."
				)
		if origin := t.get_origin(a8n):
			if origin == list:
				self.a8n_type = A8nType.LIST
				return
			elif "Union" in origin.__name__:
				self.a8n_type = A8nType.OPT
				return
		raise TypeError(
			f"Unsupported annotation '{a8n}' for "
			f"class '{cls_name}' attribute '{attr}'."
		)



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
__all__ = [
    "MultipartError",
    "NonMultipartError",
    "InvalidSeparatorError",
    "MissingContentIDError"
]


class MultipartError(Exception):
    pass


class NonMultipartError(MultipartError):
    def __init__(self, mimetype: str):
        msg = f"Unexpected mimetype in content-type header: '{mimetype}'"
        super().__init__(msg)


class InvalidSeparatorError(MultipartError):
    def __init__(self):
        msg = "Multipart request body does not conform to RFC-5322"
        super().__init__(msg)


class MissingContentIDError(MultipartError):
    def __init__(self):
        msg = f"All file attachments must have Content-ID headers"
        super().__init__(msg)



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
from .bodypart import DecodedBodyPart
from .decoder import MultipartDecoder
from .encoder import MultipartEncoder
from .errors import (
	MultipartError,
	NonMultipartError,
	InvalidSeparatorError,
	MissingContentIDError
)


__all__ = [
	"DecodedBodyPart",
	"MultipartDecoder",
	"MultipartEncoder",
	"MultipartError",
	"NonMultipartError",
	"InvalidSeparatorError",
	"MissingContentIDError"
]



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from pydantic_xml import BaseXmlModel, element, attr
from ..constants import WSDL_NSMAP
from .port_type import WSDLPortType
from .binding import WSDLBinding
from .service import WSDLService
from .schema import Schema


__all__ = [
    "WSDLTypes",
    "WSDLPart",
    "WSDLMessage",
    "WSDLDefinitions"
]


class WSDLTypes(BaseXmlModel, tag="types", nsmap=WSDL_NSMAP):
    service_schema: Schema


class WSDLPart(BaseXmlModel, tag="part", ns='wsdl', nsmap=WSDL_NSMAP):
    element: str = attr()
    name: str = attr(default="parameters")


class WSDLMessage(BaseXmlModel, tag="message", ns='wsdl', nsmap=WSDL_NSMAP):
    name: str = attr()
    parts: t.List[WSDLPart]


class WSDLDefinitions(BaseXmlModel, tag="definitions", ns="wsdl", nsmap=WSDL_NSMAP):
    target_ns: str = attr(name="targetNamespace")
    name: str = attr()
    types: WSDLTypes = element()
    messages: t.List[WSDLMessage]
    port_type: WSDLPortType
    binding: WSDLBinding
    service: WSDLService



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import quopri
import base64
import typing as t
from email.parser import HeaderParser
from email.message import Message
from .errors import (
    InvalidSeparatorError,
    MissingContentIDError
)
from .. import utils


__all__ = ["DecodedBodyPart"]


class DecodedBodyPart:
    headers: t.Optional[Message] = None
    file_name: t.Optional[str] = None
    content: t.Optional[bytes] = None
    content_id: t.Optional[str] = None
    is_mixed_multipart: bool = False

    def __init__(self, content: bytes) -> None:
        separator = b'\r\n\r\n'
        if separator not in content:
            raise InvalidSeparatorError()

        headers, content = utils.split_on_find(content, separator)
        if headers:
            decoded: str = utils.detect_decode(headers)[0]
            self.headers = HeaderParser().parsestr(decoded)

            if content:
                self.content = content
                content_type = self.headers.get_content_type()
                content_disp = self.headers.get_content_disposition()
                boundary = self.headers.get_boundary()

                if content_type and "multipart/mixed" in content_type:
                    self.is_mixed_multipart = True
                elif content_disp == "attachment" and boundary is None:
                    self.content = self.decode_transfer(content)
                    self.set_file_metadata()

    def decode_transfer(self, content: bytes) -> bytes:
        transfer_enc = self.headers.get("Content-Transfer-Encoding")
        if transfer_enc == "base64":
            return base64.b64decode(content)
        elif transfer_enc == "quoted-printable":
            return quopri.decodestring(content)
        elif transfer_enc == "binary":
            return content
        return content

    def set_file_metadata(self) -> None:
        boundary = self.headers.get_boundary()
        content_disp = self.headers.get_content_disposition()
        if boundary is None and content_disp == "attachment":
            self.file_name = self.headers.get_filename()
            if self.file_name is None:
                self.file_name = self.headers.get_param(
                    header='content-disposition',
                    param='name'
                )
            cid = self.headers.get("Content-ID")
            if cid is None:
                raise MissingContentIDError
            cid = f"cid:{cid.lstrip('<').rstrip('>')}"
            self.content_id = cid



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
from __future__ import annotations
import typing as t
from pydantic_xml import model
from pydantic import (
	PrivateAttr,
	ValidationInfo,
	model_validator
)
from ..constants import A8nType
from .meta import CompositeMeta

try:
	from .spec import BaseElementSpec
except ImportError:  # pragma: no cover
	BaseElementSpec: t.TypeAlias = t.Any


__all__ = ["MessageBody", "MessageBodyType"]


class MessageBody(model.BaseXmlModel, metaclass=CompositeMeta, search_mode='unordered', skip_empty=True):
	_element_specs: t.Dict[str, BaseElementSpec] = PrivateAttr(default_factory=dict)

	@staticmethod
	def _validate_list(attr: str, data: t.List[MessageBody], spec: BaseElementSpec):
		if not isinstance(data, list):
			raise ValueError(
				f"expected attribute '{attr}' value to be a list, "
				f"but received '{type(data).__name__}' instead."
			)
		count = len(data)
		if count < spec.min_occurs:
			raise ValueError(
				f"expected at least {spec.min_occurs} "
				f"{spec.tag} elements, got {count}"
			)
		elif isinstance(spec.max_occurs, int) and count > spec.max_occurs:
			raise ValueError(
				f"expected at most {spec.max_occurs} "
				f"{spec.tag} elements, got {count}"
			)
		expected = spec.internal_type or spec.element_type
		for item in data:
			if not isinstance(item, expected):
				raise ValueError(
					f"unexpected type '{type(item).__name__}' "
					f"in list for argument '{attr}'"
				)

	# noinspection PyNestedDecorators
	@model_validator(mode="before")
	@classmethod
	def _validate_before(cls, data: t.Dict[str, t.Any], info: ValidationInfo) -> t.Any:
		# Used for validating model instantiation in user code
		if (info.context or {}).get("deserializing", False):
			return data
		elif not isinstance(data, dict):
			return data

		private_attrs = getattr(cls, "__private_attributes__")
		class_specs = private_attrs.get("_element_specs").get_default()

		for attr, spec in class_specs.items():
			expected = spec.internal_type or spec.element_type
			value = data.get(attr)

			if spec.a8n_type == A8nType.LIST:
				cls._validate_list(attr, value, spec)
				spec.init_instantiated_data(value)
				continue
			elif spec.a8n_type == A8nType.OPT and value is None:
				data[attr] = []
				continue
			elif spec.a8n_type == A8nType.MAND and attr not in data:
				raise ValueError(f"argument '{attr}' is missing")
			elif not isinstance(value, expected):
				arg_type = type(value).__name__
				raise ValueError(
					f"unexpected type '{arg_type}' "
					f"for argument '{attr}'"
				)
			data[attr] = [value]
			spec.init_instantiated_data(data[attr])
		return data

	@model_validator(mode="after")
	def _validate_after(self, info: ValidationInfo) -> MessageBody:
		# Used for validating deserialized incoming requests
		if not (info.context or {}).get("deserializing", False):
			return self

		specs = getattr(self, "_element_specs", None)
		if not specs:
			return self

		for attr, value in vars(self).items():
			spec = specs.get(attr)  # type: BaseElementSpec
			if spec is None:
				continue
			elif not isinstance(value, list):
				value = [value]
			count = len(value)

			if count > 1 and spec.a8n_type in [A8nType.MAND, A8nType.OPT]:
				raise ValueError(f"only one {spec.tag} element is allowed")
			elif count == 0 and spec.a8n_type == A8nType.MAND:
				raise ValueError(f"must provide at least one {spec.tag} element")
			elif spec.a8n_type == A8nType.LIST:
				self._validate_list(attr, value, spec)
				spec.init_deserialized_data(value)
				continue

			spec.init_deserialized_data(value)
			value = value[0] if count == 1 else None
			setattr(self, attr, value)
		return self


MessageBodyType = t.TypeVar("MessageBodyType", bound=MessageBody)



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import inspect
import typing as t
from pathlib import Path
from lxml.etree import LxmlError
from pydantic import ValidationError
from fastapi import FastAPI, Request, Response
from fastapi.types import DecoratedCallable
from starlette.types import Lifespan
from starlette.middleware.base import BaseHTTPMiddleware
from ..multipart import MultipartError
from ..storage import GlobalWeakStorage
from .. import utils, wsdl
from .action import SoapAction
from . import faults as f


__all__ = ["SoapService"]


SoapMiddleware: t.TypeAlias = BaseHTTPMiddleware
FuncOrCoro = t.Union[t.Callable[..., t.Any], t.Awaitable[t.Any]]


class SoapService(FastAPI):
	def __init__(
			self,
			*,
			name: str = "SoapService",
			path: str = "/service",
			this_namespace: str = "https://example.org",
			wsdl_override: t.Optional[t.Union[str, Path]] = None,
			lifespan: t.Optional[Lifespan[FastAPI]] = None,
			fault_callback: t.Optional[t.Callable[[Request, Exception], None]] = None,
			hide_ise_cause: bool = False
	) -> None:
		self._name = name
		self._tns = this_namespace
		self._wsdl_response = None
		self._wsdl_override = wsdl_override
		self._fault_callback = fault_callback
		self._hide_ise_cause = hide_ise_cause
		self._storage = GlobalWeakStorage()
		self._actions = dict()

		if isinstance(wsdl_override, (str, Path)):
			wsdl_file = utils.read_cached_xml_file(wsdl_override)
			self._wsdl_response = Response(
				media_type="text/xml",
				content=wsdl_file
			)
		super().__init__(
			root_path=path,
			lifespan=lifespan,
			openapi_url=None,
			redoc_url=None,
			docs_url=None,
		)
		self._as_fastapi().add_middleware(
			middleware_class=SoapMiddleware,
			dispatch=self._soap_middleware
		)

	def _as_fastapi(self) -> FastAPI:
		return t.cast(FastAPI, self)

	async def _soap_middleware(self, http_request: Request, _: t.Callable) -> t.Optional[Response]:
		try:
			if "wsdl" in http_request.query_params:
				if self._wsdl_response is None:
					self.regenerate_wsdl()
				return self._wsdl_response
			elif http_request.method != "POST":
				raise f.InvalidMethodFault(http_request.method)

			action_name = http_request.headers.get("soapaction", '').strip('"')
			if not action_name or action_name not in self._actions.keys():
				raise f.InvalidActionFault(action_name)

			http_body = await http_request.body()
			content_type = http_request.headers.get("content-type")

			action = self._actions[action_name]
			envelope = action.parse(http_body, content_type)
			args = action.arguments_from(envelope, action_name)
			ret = await self._await_or_call(action.handler, *args)
			return action.response_from(ret, envelope.header)

		except f.SoapFault as ex:
			err, resp = ex, ex.response
		except ValidationError as ex:
			err, resp = ex, f.ValidationFault(ex).response
		except (MultipartError, LxmlError) as ex:
			err, resp = ex, f.ClientFault(ex).response
		except Exception as ex:
			if self._hide_ise_cause:
				ex = "Internal Server Error"
			err, resp = ex, f.ServerFault(ex).response

		if self._fault_callback is not None:
			await self._await_or_call(
				self._fault_callback,
				http_request, err
			)
		return resp

	@staticmethod
	async def _await_or_call(func: FuncOrCoro, *args, **kwargs) -> t.Any:
		if inspect.iscoroutinefunction(func):
			return await func(*args, **kwargs)
		return func(*args, **kwargs)

	def add_action(self, name: str, handler: t.Callable[..., t.Any]) -> None:
		if name in self._actions.keys():
			raise ValueError(f"Cannot add duplicate {name} SOAP action.")
		anno = utils.validate_annotations(name, handler)
		pos = utils.extract_parameter_positions(anno)
		self._actions[name] = SoapAction(
			name=name,
			handler=handler,
			body_type=anno.get("body"),
			body_index=pos.get("body"),
			header_type=anno.get("header"),
			header_index=pos.get("header"),
			return_type=anno.get("return"),
			storage=self._storage
		)

	def action(self, name: str) -> t.Callable[[DecoratedCallable], DecoratedCallable]:
		def closure(func: DecoratedCallable) -> DecoratedCallable:
			self.add_action(name, func)
			return func
		return closure

	def regenerate_wsdl(self, *, force: bool = False) -> None:
		if self._wsdl_override is not None and not force:
			raise RuntimeError(
				"WSDL regeneration must be explicitly forced when "
				"SoapService has wsdl_override argument set."
			)
		self._wsdl_response = Response(
			media_type="text/xml",
			content=wsdl.generate(
				self._actions,
				self._name,
				self._tns
			)
		)



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import re
import typing as t
from dataclasses import dataclass
from pydantic import ValidationError
from fastapi import Response
from ..base import MessageBody
from ..storage import GlobalWeakStorage
from ..constants import HEADER_NSMAP
from ..multipart import (
	MultipartDecoder,
	DecodedBodyPart
)
from ..envelope import (
	EnvelopeFactory,
	GenericEnvelope,
	XroadHeader
)
from .response import SoapResponse
from . import faults as f


__all__ = ["SoapAction"]


@dataclass(frozen=True)
class SoapAction:
	name: str
	handler: t.Callable[..., t.Optional[MessageBody]]
	body_type: t.Type[MessageBody]
	body_index: t.Optional[int]
	header_type: t.Type[XroadHeader]
	header_index: t.Optional[int]
	return_type: t.Optional[t.Type[MessageBody]]
	storage: GlobalWeakStorage

	def arguments_from(
			self,
			envelope: GenericEnvelope,
			action_name: str
	) -> t.List[t.Union[MessageBody, XroadHeader]]:
		args = list()
		if self.body_type is not None:
			if envelope.body is None:
				raise f.MissingBodyFault(action_name)
			args.insert(self.body_index, envelope.body.content)
		if self.header_type is not None:
			if envelope.header is None:
				raise f.MissingHeaderFault(action_name)
			args.insert(self.header_index, envelope.header)
		return args

	def response_from(
			self,
			ret_obj: t.Optional[MessageBody],
			header: XroadHeader
	) -> t.Union[Response, SoapResponse]:
		has_return = self.return_type is not None
		if not has_return and ret_obj is None:
			return Response()
		elif has_return and isinstance(ret_obj, self.return_type):
			return SoapResponse(content=ret_obj, header=header)
		raise TypeError(
			f"Expected return type {self.return_type}, "
			f"but received {ret_obj}"
		)

	def parse(self, http_body: bytes, content_type: t.Optional[str]) -> GenericEnvelope:
		body_ct = content_type.split(';')[0]
		if body_ct == "multipart/related":
			files: t.List[DecodedBodyPart] = list()
			decoder1 = MultipartDecoder(http_body, content_type)
			envelope = None
			for index, part1 in enumerate(decoder1.parts):
				if index == 0:
					envelope = part1
				elif part1.is_mixed_multipart:
					content_type = part1.headers.get('content-type')
					decoder2 = MultipartDecoder(part1.content, content_type)
					for part2 in decoder2.parts:
						files.append(part2)
				else:
					files.append(part1)
			http_body = self.process_files(envelope.content, files)
		elif body_ct not in ["text/xml", "application/xml"]:
			raise f.ClientFault(f"Invalid content type: {body_ct}")
		return self.deserialize(http_body)

	def process_files(self, xml_str: bytes, files: t.List[DecodedBodyPart]) -> bytes:
		ids = [file.content_id for file in files]
		for cid in ids:
			if ids.count(cid) > 1:
				raise f.DuplicateCIDFault(cid)
		for file in files:
			cid = file.content_id.encode()
			count = xml_str.count(cid)
			if count == 0:
				raise f.MissingCIDFault(cid.decode())
			elif count > 1:
				raise f.DuplicateCIDFault(cid.decode())
			uid = self.storage.insert_object(file).encode()
			xml_str = xml_str.replace(cid, uid)
		return xml_str

	def deserialize(self, http_body: bytes) -> GenericEnvelope:
		if self.body_type is None:
			return EnvelopeFactory().deserialize(http_body)
		try:
			factory = EnvelopeFactory[self.body_type]
			return factory().deserialize(http_body)
		except ValidationError:
			extra_nsmap = self.extract_extra_nsmap(http_body)
			name = self.body_type.__name__
			last_err = None
			for ns in extra_nsmap.keys():
				new = t.cast(t.Type[MessageBody], type(
					name, (self.body_type,), {},
					ns=ns, nsmap=extra_nsmap
				))
				attrs = self.body_type.__private_attributes__
				new.__private_attributes__ = attrs
				try:
					factory = EnvelopeFactory[new]
					return factory().deserialize(http_body)
				except ValidationError as ex:
					last_err = ex
			raise last_err

	@staticmethod
	def extract_extra_nsmap(http_body: bytes) -> t.Dict[str, str]:
		pattern = r'<.*?Envelope\s*([^>]*)>'
		match = re.search(pattern, http_body.decode(), re.IGNORECASE)
		if match is None:
			raise f.ClientFault(f"Unexpected envelope structure")
		nsmap = dict()
		for raw_ns in match.group(1).split(' '):
			key, value = raw_ns[len('xmlns:'):].split('=')  # type: str, str
			if key not in HEADER_NSMAP:
				nsmap[key] = value.strip('"')
		return nsmap



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
from .response import SoapResponse
from .service import SoapService
from .action import SoapAction
from .faults import (
	SoapFault,
	ClientFault,
	ServerFault,
	InvalidMethodFault,
	InvalidActionFault,
	MissingBodyFault,
	MissingHeaderFault,
	MissingCIDFault,
	DuplicateCIDFault,
	ValidationFault
)


__all__ = [
	"SoapResponse",
	"SoapService",
	"SoapAction",
	"SoapFault",
	"ClientFault",
	"ServerFault",
	"InvalidMethodFault",
	"InvalidActionFault",
	"MissingBodyFault",
	"MissingHeaderFault",
	"MissingCIDFault",
	"DuplicateCIDFault",
	"ValidationFault"
]



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
from .content_utils import (
	split_on_find,
	guess_mime_type,
	detect_decode,
	convert_to_utf8
)
from .path_utils import (
	search_upwards,
	resolve_relpath,
	read_cached_file,
	read_cached_xml_file
)
from .route_utils import (
	get_annotations,
	validate_annotations,
	extract_parameter_positions
)


__all__ = [
	"split_on_find",
	"guess_mime_type",
	"detect_decode",
	"search_upwards",
	"resolve_relpath",
	"read_cached_file",
	"read_cached_xml_file",
	"get_annotations",
	"validate_annotations",
	"extract_parameter_positions"
]



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import chardet
import mimetypes
import typing as t
import charset_normalizer as charset
from ..base import MessageBody, BaseElementSpec


__all__ = [
	"split_on_find",
	"guess_mime_type",
	"detect_decode",
	"convert_to_utf8"
]


_USB = t.Union[str, bytes]
_USN = t.Union[str, None]


def split_on_find(content: _USB, separator: _USB) -> t.Tuple[_USB, _USB]:
	point, sep_len = content.find(separator), len(separator)
	end = point if point != -1 else len(content) + 1
	start = end + sep_len if point != -1 else 0
	first, second = content[:end], content[start:]
	return first.strip(), second.strip()


def guess_mime_type(file_name: t.Union[str, None]) -> str:
	guess = mimetypes.guess_type(file_name)[0]
	default = "application/octet-stream"
	return guess or default


def detect_decode(string: bytes, encoding: str = 'utf-8') -> t.Tuple[_USB, _USN]:
	try:
		return string.decode(encoding), encoding
	except UnicodeDecodeError:
		if charset.is_binary(string):
			return string, None
		kwargs = dict(should_rename_legacy=True)
		match = chardet.detect(string, **kwargs)
		if match and match["confidence"] >= 0.7:  # pragma: no cover
			enc = match["encoding"]
			return string.decode(enc), enc
		return string, None  # pragma: no cover


def convert_to_utf8(string: bytes) -> bytes:
	decoded, encoding = detect_decode(string)
	if encoding not in [None, 'utf-8']:
		string = decoded.encode('utf-8')
	return string



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
from __future__ import annotations
import typing as t
from abc import ABC, abstractmethod


__all__ = [
	"FileSize",
	"FileSizeKB",
	"FileSizeMB",
	"FileSizeGB"
]


class FileSize(ABC):
	KB: t.Type[FileSizeKB]
	MB: t.Type[FileSizeMB]
	GB: t.Type[FileSizeGB]

	def __init__(self, size: int = 1) -> None:
		if size < 0:
			raise ValueError("File size cannot be a negative number")
		elif not isinstance(size, int):
			raise TypeError("File size must be an integer")
		self.size = size

	@staticmethod
	def bytes_to_iec_str(byte_count: int) -> str:
		kib, mib, gib = 1024, 1024 ** 2, 1024 ** 3
		if byte_count >= gib:
			return f"{int(byte_count // gib)}GB"
		elif byte_count >= mib:
			return f"{int(byte_count // mib)}MB"
		elif byte_count >= kib:
			return f"{int(byte_count // kib)}KB"
		return f"{byte_count}B"

	@classmethod
	@abstractmethod
	def _multiplier(cls) -> int: ...

	@abstractmethod
	def __str__(self) -> str: ...

	def __int__(self) -> int:
		return self.size * self._multiplier()

	@property
	def value(self) -> int:
		return self.__int__()


class FileSizeKB(FileSize):
	@classmethod
	def _multiplier(cls) -> int:
		return 1024

	def __str__(self) -> str:
		return f"{self.size}KB"


class FileSizeMB(FileSize):
	@classmethod
	def _multiplier(cls) -> int:
		return 1024 ** 2

	def __str__(self) -> str:
		return f"{self.size}MB"


class FileSizeGB(FileSize):
	@classmethod
	def _multiplier(cls) -> int:
		return 1024 ** 3

	def __str__(self) -> str:
		return f"{self.size}GB"


FileSize.KB = FileSizeKB
FileSize.MB = FileSizeMB
FileSize.GB = FileSizeGB



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from enum import Enum
from datetime import datetime
from .common import CommonSpecTypeA


__all__ = ["DateTimeSpec", "DateTime"]


class DateTimeSpec(CommonSpecTypeA):
	def __init__(self, **kwargs) -> None:
		super().__init__(
			element_type=datetime,
			**kwargs
		)


class DateTime:
	def __new__(
			cls,
			*,
			tag: t.Optional[str] = None,
			ns: t.Optional[str] = None,
			nsmap: t.Optional[t.Dict[str, str]] = None,
			min_occurs: int = None,
			max_occurs: t.Union[int, t.Literal["unbounded"]] = None,
			min_value: t.Optional[datetime] = None,
			max_value: t.Optional[datetime] = None,
			enumerations: t.Optional[t.Type[Enum]] = None,
			pattern: t.Optional[str] = None
	) -> t.Union[datetime, t.List[datetime]]:
		kwargs = {k: v for k, v in locals().items() if v != cls}
		return t.cast(datetime, DateTimeSpec(**kwargs))



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from fastapi.types import DecoratedCallable
from ..envelope import XroadHeader
from ..base import MessageBody


__all__ = [
	"get_annotations",
	"validate_annotations",
	"extract_parameter_positions"
]


def get_annotations(func: t.Any) -> dict:
	key: str = "__annotations__"
	anno = getattr(func, key, None)
	if anno is None and hasattr(func, "__dict__"):
		anno = func.__dict__.get(key, None)
	return anno or dict()


def validate_annotations(name: str, func: DecoratedCallable) -> dict:
	anno = get_annotations(func)
	for key, value in anno.items():
		if key not in ["body", "header", "return"]:
			raise ValueError(
				f"Parameter name '{key}' not allowed for SOAP action {name}."
				"\nOnly names 'body' and 'header' can be used for parameters."
			)
		elif key == "return":
			if value is None:
				continue
			elif not isinstance(value, type) or not issubclass(value, MessageBody):
				raise TypeError(
					f"Return type annotation of the {name} SOAP action "
					"must be either 'None' or a subclass of 'MessageBody'."
				)
		elif key == "body":
			if value == XroadHeader:
				raise TypeError(
					f"Cannot set the 'XroadHeader' class as a type annotation "
					f"to the 'body' parameter of the {name} SOAP action."
				)
			elif not isinstance(value, type) or not issubclass(value, MessageBody):
				raise TypeError(
					f"The annotation of the '{key}' parameter of the {name} "
					f"SOAP action must be a subclass of 'MessageBody'."
				)
		elif key == "header" and value != XroadHeader:
			raise ValueError(
				f"The annotation of the 'headers' parameter of the {name} "
				f"SOAP action must be the 'XroadHeaders' class."
			)
	return anno


def extract_parameter_positions(anno: dict) -> dict:
	pos = dict(body=None, header=None)
	keys = list(anno.keys())
	for key in ["body", "header"]:
		if key in keys:
			pos[key] = keys.index(key)
	return pos



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import re
import typing as t
from pydantic_xml import element
from ..base import MessageBody
from ..envelope import GenericFault
from .response import SoapResponse


__all__ = [
	"SoapFault",
	"ClientFault",
	"ServerFault",
	"InvalidMethodFault",
	"InvalidActionFault",
	"MissingBodyFault",
	"MissingHeaderFault",
	"MissingCIDFault",
	"DuplicateCIDFault",
	"ValidationFault"
]


class SoapFault(Exception):
	def __init__(
			self,
			code: str = "Client",
			string: str = "Bad Request",
			actor: t.Optional[str] = None,
			detail: t.Optional[MessageBody] = None,
			http_status_code: t.Optional[int] = 400
	) -> None:
		fault_type = MessageBody
		kwargs = dict(
			faultcode=code,
			faultstring=string,
			faultactor=actor
		)
		if isinstance(detail, MessageBody):
			fault_type = detail.__class__
			kwargs["detail"] = detail

		typed_fault = GenericFault[fault_type]
		fault = typed_fault(**kwargs)

		self.response = SoapResponse(
			http_status_code=http_status_code,
			content=fault,
			header=None
		)


class ClientFault(SoapFault):
	def __init__(self, ex: t.Union[str, Exception]) -> None:
		super().__init__(string=str(ex))


class ServerFault(SoapFault):
	def __init__(self, ex: t.Union[str, Exception]) -> None:
		super().__init__(
			http_status_code=500,
			string=str(ex),
			code="Server"
		)


class InvalidMethodFault(SoapFault):
	def __init__(self, method: str) -> None:
		msg = f"Service does not support {method} method for SOAP action requests"
		super().__init__(string=msg)


class InvalidActionFault(SoapFault):
	def __init__(self, action: str) -> None:
		msg = f"Service does not support SOAP action: {action}"
		super().__init__(string=msg)


class MissingBodyFault(SoapFault):
	def __init__(self, action_name: str) -> None:
		msg = f"Body element missing from envelope for {action_name} SOAP action"
		super().__init__(string=msg)


class MissingHeaderFault(SoapFault):
	def __init__(self, action_name: str) -> None:
		msg = f"X-Road header element missing from envelope for {action_name} SOAP action"
		super().__init__(string=msg)


class MissingCIDFault(SoapFault):
	def __init__(self, cid: str) -> None:
		msg = f"Content-ID missing from envelope: {cid}"
		super().__init__(string=msg)


class DuplicateCIDFault(SoapFault):
	def __init__(self, cid: str) -> None:
		msg = f"Duplicate Content-ID not allowed: {cid[4:]}"
		super().__init__(string=msg)


class ValidationFault(SoapFault):
	def __init__(self, ex: Exception) -> None:
		error = str(ex)
		match = re.search(r"^(.*?)\sfor", error)
		string = match.group(1) if match else "validation error"
		super().__init__(
			http_status_code=400,
			detail=self.extract_details(error),
			string=string + " in SOAP envelope",
			code="Client"
		)

	@staticmethod
	def extract_details(error: str) -> MessageBody:
		class Detail(MessageBody):
			location: str = element()
			reason: str = element()
			input_value: str = element()

		class ErrorDetails(MessageBody):
			details: t.List[Detail] = element(tag="validationError")

		details = list()
		pattern = r"(body\..+?)(?=\nbody\.|\Z)"
		matches = re.findall(pattern, error, re.DOTALL)

		for match in matches:
			loc, msg = match.split('\n')
			iv_match = re.search(r"\$\$(.*?)\$\$", msg)
			if iv_match:
				msg = msg.replace(f"$${iv_match.group(1)}$$", '')
			else:
				iv_match = re.search(r"input_value=(.+?),", msg)
			ln_match = re.search(r"(\[line -?\d+]: )", msg)
			re_match = re.search(r"\[line -?\d+]: (.+?) \[type=", msg)
			details.append(Detail(
				location=(ln_match.group(1) + loc) if ln_match else "unknown",
				input_value=iv_match.group(1) if iv_match else "unknown",
				reason=re_match.group(1) if re_match else "unknown"
			))
		return ErrorDetails(details=details)



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from pydantic_xml import BaseXmlModel, attr
from .conditions import (
	Enumeration,
	RegexPattern,
	MinInclusive,
	MaxInclusive,
	MinExclusive,
	MaxExclusive,
	FractionDigits,
	TotalDigits,
	Length,
	MinLength,
	MaxLength,
	WhiteSpace
)


__all__ = [
	"StringRestriction",
	"AnyURIRestriction",
	"IntegerRestriction",
	"DecimalRestriction",
	"FloatRestriction",
	"DoubleRestriction",
	"DateRestriction",
	"TimeRestriction",
	"DateTimeRestriction",
	"DurationRestriction"
]


class StringRestriction(BaseXmlModel, tag="restriction"):
	base: str = attr(default="string")
	length: t.Union[Length, None] = None
	min_length: t.Union[MinLength, None] = None
	max_length: t.Union[MaxLength, None] = None
	enumerations: t.Union[t.List[Enumeration], None] = None
	whiteSpace: t.Union[WhiteSpace, None] = None
	pattern: t.Union[RegexPattern, None] = None


class AnyURIRestriction(BaseXmlModel, tag="restriction"):
	base: str = attr(default="anyURI")
	length: t.Union[Length, None] = None
	min_length: t.Union[MinLength, None] = None
	max_length: t.Union[MaxLength, None] = None
	enumerations: t.Union[t.List[Enumeration], None] = None
	whiteSpace: t.Union[WhiteSpace, None] = None
	pattern: t.Union[RegexPattern, None] = None


class IntegerRestriction(BaseXmlModel, tag="restriction"):
	base: str = attr(default="integer")
	min_inclusive: t.Union[MinInclusive, None] = None
	max_inclusive: t.Union[MaxInclusive, None] = None
	min_exclusive: t.Union[MinExclusive, None] = None
	max_exclusive: t.Union[MaxExclusive, None] = None
	total_digits: t.Union[TotalDigits, None] = None
	enumerations: t.Union[t.List[Enumeration], None] = None
	pattern: t.Union[RegexPattern, None] = None


class DecimalRestriction(BaseXmlModel, tag="restriction"):
	base: str = attr(default="decimal")
	min_inclusive: t.Union[MinInclusive, None] = None
	max_inclusive: t.Union[MaxInclusive, None] = None
	min_exclusive: t.Union[MinExclusive, None] = None
	max_exclusive: t.Union[MaxExclusive, None] = None
	total_digits: t.Union[TotalDigits, None] = None
	fraction_digits: t.Union[FractionDigits, None] = None
	enumerations: t.Union[t.List[Enumeration], None] = None
	pattern: t.Union[RegexPattern, None] = None


class FloatRestriction(BaseXmlModel, tag="restriction"):
	base: str = attr(default="float")
	min_inclusive: t.Union[MinInclusive, None] = None
	max_inclusive: t.Union[MaxInclusive, None] = None
	min_exclusive: t.Union[MinExclusive, None] = None
	max_exclusive: t.Union[MaxExclusive, None] = None
	enumerations: t.Union[t.List[Enumeration], None] = None
	pattern: t.Union[RegexPattern, None] = None


class DoubleRestriction(BaseXmlModel, tag="restriction"):
	base: str = attr(default="double")
	min_inclusive: t.Union[MinInclusive, None] = None
	max_inclusive: t.Union[MaxInclusive, None] = None
	min_exclusive: t.Union[MinExclusive, None] = None
	max_exclusive: t.Union[MaxExclusive, None] = None
	enumerations: t.Union[t.List[Enumeration], None] = None
	pattern: t.Union[RegexPattern, None] = None


class DateRestriction(BaseXmlModel, tag="restriction"):
	base: str = attr(default="date")
	min_inclusive: t.Union[MinInclusive, None] = None
	max_inclusive: t.Union[MaxInclusive, None] = None
	min_exclusive: t.Union[MinExclusive, None] = None
	max_exclusive: t.Union[MaxExclusive, None] = None
	enumerations: t.Union[t.List[Enumeration], None] = None
	pattern: t.Union[RegexPattern, None] = None


class TimeRestriction(BaseXmlModel, tag="restriction"):
	base: str = attr(default="time")
	min_inclusive: t.Union[MinInclusive, None] = None
	max_inclusive: t.Union[MaxInclusive, None] = None
	min_exclusive: t.Union[MinExclusive, None] = None
	max_exclusive: t.Union[MaxExclusive, None] = None
	enumerations: t.Union[t.List[Enumeration], None] = None
	pattern: t.Union[RegexPattern, None] = None


class DateTimeRestriction(BaseXmlModel, tag="restriction"):
	base: str = attr(default="dateTime")
	min_inclusive: t.Union[MinInclusive, None] = None
	max_inclusive: t.Union[MaxInclusive, None] = None
	min_exclusive: t.Union[MinExclusive, None] = None
	max_exclusive: t.Union[MaxExclusive, None] = None
	enumerations: t.Union[t.List[Enumeration], None] = None
	pattern: t.Union[RegexPattern, None] = None


class DurationRestriction(BaseXmlModel, tag="restriction"):
	base: str = attr(default="duration")
	min_inclusive: t.Union[MinInclusive, None] = None
	max_inclusive: t.Union[MaxInclusive, None] = None
	min_exclusive: t.Union[MinExclusive, None] = None
	max_exclusive: t.Union[MaxExclusive, None] = None
	enumerations: t.Union[t.List[Enumeration], None] = None
	pattern: t.Union[RegexPattern, None] = None



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import secrets


__all__ = ["CIDGenerator"]


class CIDGenerator:
	def __init__(self, max_len: int = 30):
		self.items = list()
		self._max_len = max_len
		self._token_len = 20

	@property
	def token(self) -> str:
		exists_counter = 0

		while True:
			if self._token_len > self._max_len:
				raise RuntimeError("CIDGenerator out of bounds")

			size = self._token_len // 2
			token = secrets.token_hex(size)
			cid = f"cid:{token}"

			if cid not in self.items:
				self.items.append(cid)
				return cid
			elif exists_counter < 10:
				exists_counter += 1
			else:
				self._token_len += 2
				exists_counter = 0
